#ifndef _INIFILE__H
#define _INIFILE__H
#endif


//
// ini.h
//
// This class is used to read .ini files and obtain the information from
// these files. It will load the complete file into memory, and then work
// out the various sections, and the section items and any values associated
// with those items. It can be initialised to be either case insensitive or
// case sensitive.
//

/*
	Here are some ways to use the functions:
	Get a particular value:
	char *p;
	p = i_f->GetValue("Preferences", "auto_select");

	Print out a list of all section headers
	char **p; int i;
	i = i_f->NumberOfSections();
	p = i_f->SectionNames();

	for (int j=0; j<i; j++)
		printf("%s|\n", p[j]);

*/
		

#include <stdio.h>

#define _TRUE_ 1
#define _FALSE_ 0

class IniEntry
{
	private:
		int first_non_space;
		int equals_position;
		int last_letter;
		int first_after_equals;
		int last_after_equals;

	public:
		IniEntry(char **fileimage, int first_line, int last_line);
		~IniEntry();
		char* EntryValue(char *entry);
		char** EntryNames();
		int NumberOfEntries();
		
	protected:
		int number_of_entries;
		char **entry_names;		// the names of the entries in the section
		char **entry_values;	// the values of each entry

	private:
		char HasEntryName(char *textline);
		char ContainsEqualsSign(char *textline);
		void LastLetterBeforeEquals(char *textline);
		void AddEntryName(char *textline, int entry_number);
		void AddEntryValue(char *textline, int entry_number);
		char FirstLetterAfterEquals(char *textline);
		void LastLetterAfterEquals(char *textline);
		void RemoveTrailingSpaces(char *textline);
		int FindEntry(char *entry);

};


class IniFile
{
	private:
		int open_bracket_pos;
		int close_bracket_pos;

	public:
		IniFile(char *filename);
		virtual ~IniFile();
		char* GetValue(char *section, char *entry);
		char** SectionNames();
		char** SectionEntries(char *section);
		int NumberOfSections();
		int NumberOfEntries(char *section);
		
	protected:
		char ini_file_read;			// whether the ini file was opened or not
		int number_of_lines;		// the number of lines in the .ini file
		int number_of_sections;	// the number of individual sections in the .ini file
		
		char **inifile;					// copy of the .ini file
		char **section_names;		// the names of the sections in the file
		long int *section_pos;	// the lines of the file the sections start on
		IniEntry **entries;			// the entries for each section

	private:
		char FirstLetterIsOpenBracket(int line);
		char ContainsCloseBracket(int line);
		void CountNumberOfSections();
		void GetSectionNames();
		int FindSection(char *section);
};



